"""
A module to define types used by recon3d.

This module defines various custom types used throughout the recon3d application.
These types provide clear and consistent data structures for different components
of the application.

Types
-----
CartesianAxis2D
    Constant axis ordering in Python for 2D arrays (Y, X) ordering.
CartesianAxis3D
    Constant axis ordering in Python for 3D arrays (Z, Y, X) ordering.
DataVolume
    The integer counts of the data that compose a single image or a volume composed of a stack of images.
Units
    Unit enum for metadata.
UnitVector
    Normalized unit vector.
Angle
    Angle class with associated units.
Area
    Area class with associated units.
InstanceLabels
    Collection of instance labels.
Length
    Length class with associated units.
NthNearestNeighbors
    Nth nearest neighbors for instance analysis.
NVoxel
    Number of voxels, typically for an instance object.
Volume
    Volume class with associated units.
Centroid
    The centroid location for the X, Y, and Z axes, in length units.
Centroids
    Centroids for an instance image stack.
EllipsoidAxis
    Properties of best fit ellipsoid semi-axis.
EquivalentSphereDiameters
    Collection of spherical diameters.
Origin
    The origin in spatial coordinates.
Resolution
    The voxel resolution for the X, Y, and Z dimension in length units.
BestFitEllipsoid
    Three semi-axes of best-fit ellipsoid in descending order of size.
BestFitEllipsoids
    Collection of best-fit ellipsoids.
EllipsoidSurfaceAreas
    Collection of ellipsoid surface area objects.
EllipsoidVolumes
    Collection of ellipsoid volumes.
MetaData
    The data description of the scan.
ImageStack
    The generic image stack.
InstanceImageStack
    The image stack where each pixel in the image is categorized (labeled) into a specific instance of a class or object.
InstanceIndices
    Labels and index locations for each instance in an InstanceImageStack.
SemanticImageStack
    The image stack where each pixel in the image is categorized (labeled) into a class or object.
VoidDescriptorEllipsoidAngles
    Angles for ellipsoid properties to be used in void descriptor function analysis.
InstanceProperties
    Properties for labeled features in an instance image stack.
VoidDescriptorFittingParameters
    Fitting parameters for void descriptor function analysis.
VoidDescriptorProperties
    Properties for calculation of void descriptor function.
VoidDescriptorPoreProperties
    Properties for void descriptor pore analysis.

Examples
--------
To use `CartesianAxis2D` in your code:

    >>> from custom_types_module import CartesianAxis2D
    >>> axis = CartesianAxis2D.X
    >>> print(axis)
    CartesianAxis2D.X
"""

from typing import NamedTuple
from enum import Enum, IntEnum
import numpy as np
import numpy.typing as npt

##### BASE CLASSES #####


class CartesianAxis2D(IntEnum):
    """
    Constant axis ordering in Python for 2D arrays (Y, X) ordering.

    The 2D numpy arrays are ordered Y, X.

    Attributes
    ----------
    X : int
        The X-axis (1).
    Y : int
        The Y-axis (0).

    Examples
    --------
    >>> axis = CartesianAxis2D.X
    >>> print(axis)
    CartesianAxis2D.X
    """

    X = 1
    Y = 0


class CartesianAxis3D(IntEnum):
    """
    Constant axis ordering in Python for 3D arrays (Z, Y, X) ordering.

    The 3D numpy arrays are ordered Z, Y, X.

    Attributes
    ----------
    X : int
        The X-axis (2).
    Y : int
        The Y-axis (1).
    Z : int
        The Z-axis (0).

    Examples
    --------
    >>> axis = CartesianAxis3D.Z
    >>> print(axis)
    CartesianAxis3D.Z
    """

    X = 2
    Y = 1
    Z = 0


class DataVolume(NamedTuple):
    """
    The integer counts of the data that compose a single image or a volume composed of a stack of images.

    For the 2D case, the units are pixels and the z_image_count=1.
    For the 3D case, the units are voxels, and the z_image_count>=2.

    Attributes
    ----------
    x_width : int
        The width of the data volume in the X dimension.
    y_height : int
        The height of the data volume in the Y dimension.
    z_image_count : int
        The number of images in the Z dimension.

    Examples
    --------
    >>> volume = DataVolume(x_width=8, y_height=3, z_image_count=2)
    >>> print(volume)
    DataVolume(x_width=8, y_height=3, z_image_count=2)
    """

    x_width: int
    y_height: int
    z_image_count: int


class Units(Enum):
    """
    Unit enum for metadata.

    Attributes
    ----------
    NANOMETER : str
        Nanometer unit.
    MICRON : str
        Micron unit.
    MILLIMETER : str
        Millimeter unit.
    METER : str
        Meter unit.
    VOXEL : str
        Voxel unit.
    PIXEL : str
        Pixel unit.
    RADIAN : str
        Radian unit.
    DEGREE : str
        Degree unit.

    Examples
    --------
    >>> unit = Units.MICRON
    >>> print(unit)
    <Units.MICRON: 'micron'>
    """

    NANOMETER: str = "nanometer"
    MICRON: str = "micron"
    MILLIMETER: str = "millimeter"
    METER: str = "meter"
    VOXEL: str = "voxel"
    PIXEL: str = "pixel"
    RADIAN: str = "radian"
    DEGREE: str = "degree"


class UnitVector(NamedTuple):
    """
    Normalized unit vector.

    Attributes
    ----------
    u : float
        The U component of the unit vector.
    v : float
        The V component of the unit vector.
    w : float
        The W component of the unit vector.

    Examples
    --------
    >>> vector = UnitVector(u=1.0, v=0.0, w=0.0)
    >>> print(vector)
    UnitVector(u=1.0, v=0.0, w=0.0)
    """

    u: float
    v: float
    w: float


##### DERIVED CLASSES #####


class Angle(NamedTuple):
    """
    Angle class with associated units.

    Attributes
    ----------
    value : float
        The value of the angle.
    unit : Units
        The unit of the angle.

    Examples
    --------
    >>> angle = Angle(value=90.0, unit=Units.DEGREE)
    >>> print(angle)
    Angle(value=90.0, unit=<Units.DEGREE: 'degree'>)
    """

    value: float
    unit: Units


class Area(NamedTuple):
    """
    Area class with associated units.

    Attributes
    ----------
    value : float
        The value of the area.
    unit_squared : Units
        The unit of the area squared.

    Examples
    --------
    >>> area = Area(value=100.0, unit_squared=Units.MICRON)
    >>> print(area)
    Area(value=100.0, unit_squared=<Units.MICRON: 'micron'>)
    """

    value: float
    unit_squared: Units


class InstanceLabels(NamedTuple):
    """
    Collection of instance labels.

    Attributes
    ----------
    data : np.ndarray[int]
        The array of instance labels.

    Examples
    --------
    >>> labels = InstanceLabels(data=np.array([1, 2, 3]))
    >>> print(labels)
    InstanceLabels(data=array([1, 2, 3]))
    """

    data: np.ndarray[int]


class Length(NamedTuple):
    """
    Length class with associated units.

    Attributes
    ----------
    value : float
        The value of the length.
    unit : Units
        The unit of the length.

    Examples
    --------
    >>> length = Length(value=10.0, unit=Units.MICRON)
    >>> print(length)
    Length(value=10.0, unit=<Units.MICRON: 'micron'>)
    """

    value: float
    unit: Units


class NthNearestNeighbors(NamedTuple):
    """
    Nth nearest neighbors for instance analysis.

    1st nearest neighbors are the features themselves.
    Distances are given in the same length units as centroids.

    Attributes
    ----------
    nth_nearest : int
        The nth nearest neighbor.
    distances : list[Length]
        The list of distances to the nth nearest neighbors.
    instance_id : np.ndarray[int]
        The array of instance IDs.

    Examples
    --------
    >>> neighbors = NthNearestNeighbors(nth_nearest=2, distances=[Length(value=5.0, unit=Units.MICRON)], instance_id=np.array([1, 2, 3]))
    >>> print(neighbors)
    NthNearestNeighbors(nth_nearest=2, distances=[Length(value=5.0, unit=<Units.MICRON: 'micron'>)], instance_id=array([1, 2, 3]))
    """

    nth_nearest: int
    distances: list[Length]
    instance_id: np.ndarray[int]


class NVoxel(NamedTuple):
    """
    Number of voxels, typically for an instance object.

    Attributes
    ----------
    value : int
        The number of voxels.
    unit : Units
        The unit of the voxels (default is Units.VOXEL).

    Examples
    --------
    >>> n_voxel = NVoxel(value=100)
    >>> print(n_voxel)
    NVoxel(value=100, unit=<Units.VOXEL: 'voxel'>)
    """

    value: int
    unit: Units = Units.VOXEL


class Volume(NamedTuple):
    """
    Volume class with associated units.

    Attributes
    ----------
    value : float
        The value of the volume.
    unit_cubed : Units
        The unit of the volume cubed.

    Examples
    --------
    >>> volume = Volume(value=1000.0, unit_cubed=Units.MICRON)
    >>> print(volume)
    Volume(value=1000.0, unit_cubed=<Units.MICRON: 'micron'>)
    """

    value: float
    unit_cubed: Units


##### DOUBLY DERIVED CLASSES #####


class Centroid(NamedTuple):
    """
    The centroid location for the X, Y, and Z axes, in length units.

    Attributes
    ----------
    cx : Length
        The centroid location in the X axis.
    cy : Length
        The centroid location in the Y axis.
    cz : Length
        The centroid location in the Z axis.

    Examples
    --------
    >>> centroid = Centroid(cx=Length(value=1.0, unit=Units.MICRON), cy=Length(value=2.0, unit=Units.MICRON), cz=Length(value=3.0, unit=Units.MICRON))
    >>> print(centroid)
    Centroid(cx=Length(value=1.0, unit=<Units.MICRON: 'micron'>), cy=Length(value=2.0, unit=<Units.MICRON: 'micron'>), cz=Length(value=3.0, unit=<Units.MICRON: 'micron'>))
    """

    cx: Length
    cy: Length
    cz: Length


class Centroids(NamedTuple):
    """
    Centroids for an instance image stack.

    Attributes
    ----------
    data : list[Centroid]
        The list of centroids.

    Examples
    --------
    >>> centroids = Centroids(data=[Centroid(cx=Length(value=1.0, unit=Units.MICRON), cy=Length(value=2.0, unit=Units.MICRON), cz=Length(value=3.0, unit=Units.MICRON))])
    >>> print(centroids)
    Centroids(data=[Centroid(cx=Length(value=1.0, unit=<Units.MICRON: 'micron'>), cy=Length(value=2.0, unit=<Units.MICRON: 'micron'>), cz=Length(value=3.0, unit=<Units.MICRON: 'micron'>))])
    """

    data: list[Centroid]


class EllipsoidAxis(NamedTuple):
    """
    Properties of best fit ellipsoid semi-axis (half of length of principal axis).

    Attributes
    ----------
    length : Length
        The length of the semi-axis.
    orientation : UnitVector
        The orientation of the semi-axis.

    Examples
    --------
    >>> axis = EllipsoidAxis(length=Length(value=5.0, unit=Units.MICRON), orientation=UnitVector(u=1.0, v=0.0, w=0.0))
    >>> print(axis)
    EllipsoidAxis(length=Length(value=5.0, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=1.0, v=0.0, w=0.0))
    """

    length: Length
    orientation: UnitVector


class EquivalentSphereDiameters(NamedTuple):
    """
    Collection of spherical diameters.

    Attributes
    ----------
    data : list[Length]
        The list of spherical diameters.

    Examples
    --------
    >>> diameters = EquivalentSphereDiameters(data=[Length(value=10.0, unit=Units.MICRON)])
    >>> print(diameters)
    EquivalentSphereDiameters(data=[Length(value=10.0, unit=<Units.MICRON: 'micron'>)])
    """

    data: list[Length]


class Origin(NamedTuple):
    """
    The origin in spatial coordinates.

    Attributes
    ----------
    x0 : Length
        The origin in the X axis.
    y0 : Length
        The origin in the Y axis.
    z0 : Length
        The origin in the Z axis.

    Examples
    --------
    >>> origin = Origin(x0=Length(value=0.0, unit=Units.MICRON), y0=Length(value=0.0, unit=Units.MICRON), z0=Length(value=0.0, unit=Units.MICRON))
    >>> print(origin)
    Origin(x0=Length(value=0.0, unit=<Units.MICRON: 'micron'>), y0=Length(value=0.0, unit=<Units.MICRON: 'micron'>), z0=Length(value=0.0, unit=<Units.MICRON: 'micron'>))
    """

    x0: Length
    y0: Length
    z0: Length


class Resolution(NamedTuple):
    """
    The voxel resolution for the X, Y, and Z dimension in length units.

    Attributes
    ----------
    dx : Length
        The resolution in the X dimension.
    dy : Length
        The resolution in the Y dimension.
    dz : Length
        The resolution in the Z dimension.

    Examples
    --------
    >>> resolution = Resolution(dx=Length(value=1.0, unit=Units.MICRON), dy=Length(value=1.0, unit=Units.MICRON), dz=Length(value=1.0, unit=Units.MICRON))
    >>> print(resolution)
    Resolution(dx=Length(value=1.0, unit=<Units.MICRON: 'micron'>), dy=Length(value=1.0, unit=<Units.MICRON: 'micron'>), dz=Length(value=1.0, unit=<Units.MICRON: 'micron'>))
    """

    dx: Length
    dy: Length
    dz: Length


##### TRIPLY DERIVED CLASSES #####


class BestFitEllipsoid(NamedTuple):
    """
    Three semi-axes of best-fit ellipsoid in descending order of size.

    Attributes
    ----------
    a : EllipsoidAxis
        The largest semi-axis.
    b : EllipsoidAxis
        The middle semi-axis.
    c : EllipsoidAxis
        The smallest semi-axis.
    """

    a: EllipsoidAxis
    b: EllipsoidAxis
    c: EllipsoidAxis


class BestFitEllipsoids(NamedTuple):
    """
    Collection of best-fit ellipsoids.

    Attributes
    ----------
    data : list[BestFitEllipsoid]
        The list of best-fit ellipsoids.
    """

    data: list[BestFitEllipsoid]


class EllipsoidSurfaceAreas(NamedTuple):
    """
    Collection of ellipsoid surface area objects.

    Attributes
    ----------
    data : list[Area]
        The list of ellipsoid surface areas.
    """

    data: list[Area]


class EllipsoidVolumes(NamedTuple):
    """
    Collection of ellipsoid volumes.

    Attributes
    ----------
    data : list[Volume]
        The list of ellipsoid volumes.
    """

    data: list[Volume]


class MetaData(NamedTuple):
    """
    The data description of the scan.

    Attributes
    ----------
    data_volume : DataVolume
        The dimensions of the volume.
    resolution : Resolution
        The voxel resolution for the X, Y, and Z dimensions.
    pixel_units : str
        The units of the pixels (e.g., micrometer).
    origin : Origin
        The origin in spatial coordinates.
    """

    data_volume: DataVolume
    resolution: Resolution
    pixel_units: str
    origin: Origin


class ImageStack(NamedTuple):
    """
    The generic image stack.

    Attributes
    ----------
    name : str
        The name of the image stack, used to create the path in the HDF5 file.
    metadata : MetaData
        The metadata associated with the image stack.
    data : npt.NDArray
        The image data.
    """

    name: str
    metadata: MetaData
    data: npt.NDArray


class InstanceImageStack(NamedTuple):
    """
    The image stack where each pixel in the image is categorized (labeled) into a specific instance of a class or object.

    Attributes
    ----------
    name : str
        The name of the image stack, used to create the path in the HDF5 file.
    metadata : MetaData
        The metadata associated with the image stack.
    data : npt.NDArray[np.int_]
        The labeled image data.
    nlabels : int
        The number of labels.
    min_feature_size : int
        The minimum feature size.
    """

    name: str
    metadata: MetaData
    data: npt.NDArray[np.int_]
    nlabels: int
    min_feature_size: int


class InstanceIndices(NamedTuple):
    """
    Labels and index locations for each instance in an InstanceImageStack.

    Attributes
    ----------
    source_name : str
        The name of the source InstanceImageStack.
    labels : InstanceLabels
        The instance labels.
    indices : list[np.ndarray[int]]
        The list of index locations for each instance.
    """

    source_name: str
    labels: InstanceLabels
    indices: list[np.ndarray[int]]


class SemanticImageStack(NamedTuple):
    """
    The image stack where each pixel in the image is categorized (labeled) into a class or object.

    Attributes
    ----------
    name : str
        The name of the image stack, used to create the path in the HDF5 file.
    metadata : MetaData
        The metadata associated with the image stack.
    data : npt.NDArray[np.int_]
        The segmented image data, of type int.
    """

    name: str
    metadata: MetaData
    data: npt.NDArray[np.int_]


class VoidDescriptorEllipsoidAngles(NamedTuple):
    """
    Angles for ellipsoid properties to be used in void descriptor function analysis.

    Attributes
    ----------
    theta_xy : Angle
        The angle in the XY plane.
    phi_x : Angle
        The angle in the X direction.
    phi_y : Angle
        The angle in the Y direction.
    """

    theta_xy: Angle
    phi_x: Angle
    phi_y: Angle


class InstanceProperties(NamedTuple):
    """
    Properties for labeled features in an instance image stack.

    Attributes
    ----------
    source_name : str
        The name of the source InstanceImageStack.
    labels : InstanceLabels
        The instance labels.
    n_voxels : list[NVoxel]
        The number of voxels for each instance.
    equivalent_sphere_diameters : list[Length]
        The equivalent sphere diameters for each instance.
    centroids : Centroids
        The centroids for each instance.
    ellipsoids : BestFitEllipsoids
        The best-fit ellipsoids for each instance.
    surface_areas : EllipsoidSurfaceAreas
        The surface areas of the ellipsoids for each instance.
    volumes : EllipsoidVolumes
        The volumes of the ellipsoids for each instance.
    """

    source_name: str
    labels: InstanceLabels
    n_voxels: list[NVoxel]
    equivalent_sphere_diameters: list[Length]
    centroids: Centroids
    ellipsoids: BestFitEllipsoids
    surface_areas: EllipsoidSurfaceAreas
    volumes: EllipsoidVolumes


class VoidDescriptorFittingParameters(NamedTuple):
    """
    Fitting parameters for void descriptor function analysis.

    Attributes
    ----------
    alpha : float
        The alpha fitting parameter.
    rho : float
        The rho fitting parameter.
    gamma : float
        The gamma fitting parameter.
    zeta : float
        The zeta fitting parameter.
    """

    alpha: float
    rho: float
    gamma: float
    zeta: float


class VoidDescriptorProperties(NamedTuple):
    """
    Properties for calculation of void descriptor function.

    Attributes
    ----------
    fitting_parameters : VoidDescriptorFittingParameters
        The fitting parameters for the void descriptor function.
    metadata : MetaData
        The metadata associated with the void descriptor properties.
    specimen_volume : Volume
        The volume of the specimen.
    loading_direction : CartesianAxis3D
        The loading direction.
    max_radial_distance : Length
        The maximum radial distance from the center to the free surface.
    gauge_length : Length
        The gauge length.
    """

    fitting_parameters: VoidDescriptorFittingParameters
    metadata: MetaData
    specimen_volume: Volume
    loading_direction: CartesianAxis3D
    max_radial_distance: Length
    gauge_length: Length


class VoidDescriptorPoreProperties(NamedTuple):
    """
    Properties for void descriptor pore analysis.

    Attributes
    ----------
    centroid_px : Centroid
        The centroid of the pore.
    volume : Volume
        The volume of the pore.
    surface_area : Area
        The surface area of the pore.
    r_a : Length
        The semi-major axis length of the pore.
    r_b : Length
        The semi-intermediate axis length of the pore.
    r_c : Length
        The semi-minor axis length of the pore.
    vdf_angles : VoidDescriptorEllipsoidAngles
        The angles for the ellipsoid properties.
    """

    centroid_px: Centroid
    volume: Volume
    surface_area: Area
    r_a: Length
    r_b: Length
    r_c: Length
    vdf_angles: VoidDescriptorEllipsoidAngles



# This is a blank file used to allow import from subdirectories.
# References:
# https://stackoverflow.com/questions/1260792/import-a-file-from-a-subdirectory
# https://packaging.python.org/tutorials/packaging-projects/#a-simple-project

# Add import pdbp to an __init__.py of your project, which
# will automatically make Pdb+ the default debugger at breakpoints
# import pdbp



"""
HDF5 Processing Module
=======================

This module provides a set of functions for creating, modifying, and interacting with HDF5 files,
as well as converting between image stacks and voxel data. It includes command line interfaces
for some of the key functionalities.

Functions
---------
create_hdf(hdf_path, base_container)
    Prescribe foundational structure of the HDF5 file.

modify_hdf_dataset(hdf_path, dataset_loc, data, dtype, operation, mutable)
    Modify an existing HDF5 file to alter data within a container.

write_attr_dict(hdf_path, d, dataset_loc)
    Write attributes to a specified dataset in the HDF5 file.

write_h5(hdf_path, data)
    Write data to an HDF5 file.

add_to_h5(data, hdf_path, hdf_group)
    Add data to an HDF5 file based on its type.

image_to_voxel(yml_path)
    Populate the HDF5 file with the semantic segmentation image stack specified in the YAML file, including metadata.

image_to_voxel_command_line()
    The command line wrapper for the `image_to_voxel` function.

voxel_to_image(yml_path)
    Save the image data within the HDF5 file as TIFFs in a new directory.

voxel_to_image_command_line()
    The command line wrapper for the `voxel_to_image` function.

Examples
--------
To create an HDF5 file with a base container:

    >>> create_hdf(Path("output.h5"), "base_group")

To modify an HDF5 dataset:

    >>> data = np.random.rand(10, 5)
    >>> modify_hdf_dataset(Path("output.h5"), "group/dataset", data, np.float64, "create", True)

To add data to an HDF5 file based on its type:

    >>> centroids = Centroids(...)
    >>> add_to_h5(centroids, Path("output.h5"), "group_name")

To populate an HDF5 file with image stack data from a YAML file:

    >>> yml_path = Path("config.yml")
    >>> hdf5_path = image_to_voxel(yml_path)

To convert voxel data in an HDF5 file to image files:

    >>> yml_path = Path("config.yml")
    >>> image_dir = voxel_to_image(yml_path)

To run the `image_to_voxel` function from the command line:

    $ python -m your_module_name input_file.yml

To run the `voxel_to_image` function from the command line:

    $ python -m your_module_name input_file.yml
"""

import argparse
from pathlib import Path
import numpy as np
import h5py
from functools import singledispatch

# from recon3d.feature_analysis import SemanticImageStack
from recon3d.types import *
import recon3d.utility as ut
import recon3d.instance_analysis as ia
import recon3d.types as cs


### BASE/HELPER FUNCTIONS ###
def create_hdf(hdf_path: Path, base_container: str) -> bool:
    """
    Prescribe foundational structure of the HDF5 file.

    This function creates an HDF5 file at the specified path and initializes it with a base container group.

    Parameters
    ----------
    hdf_path : Path
        The path to the location of the HDF5 file.
    base_container : str
        The name of the base container group in the HDF5 file.

    Returns
    -------
    bool
        True if the HDF5 file is created successfully, False otherwise.

    Examples
    --------
    >>> create_hdf(Path("output.h5"), "base_group")
    True
    """

    with h5py.File(hdf_path, "w") as file:
        file.create_group(base_container)


def modify_hdf_dataset(
    hdf_path: Path,
    dataset_loc: str,
    data: np.ndarray,
    dtype: type,
    operation: str,
    mutable: bool,
) -> bool:
    """
    Modify an existing HDF5 file to alter data within a container.

    This function modifies an HDF5 file by creating, appending, or overwriting a dataset at the specified location.

    Parameters
    ----------
    hdf_path : Path
        The path to the location of the HDF5 file.
    dataset_loc : str
        The internal path to the dataset in the HDF5 file (e.g., "file/container/container2/dataset").
    data : np.ndarray
        The array to be written to the HDF5 file at the specified dataset location.
    dtype : type
        The data type of the dataset (e.g., np.float64, np.int, np.uint16, for string: h5py.special_dtype(vlen=str)).
    operation : str
        The operation to perform on the dataset:
        - "create": create a new dataset
        - "append": append to an existing dataset along dimension 0
        - "overwrite": overwrite an existing dataset along dimension 0 (e.g., shrinking dataset)
    mutable : bool
        If True, the initial shape is zero in the first dimension, with no size limit.

    Returns
    -------
    bool
        True if the operation is successful, False otherwise.

    Raises
    ------
    ValueError
        If the dataset already exists when trying to create a new one.
    KeyError
        If an unsupported operation is requested.

    Examples
    --------
    >>> data = np.random.rand(10, 5)
    >>> modify_hdf_dataset(Path("output.h5"), "group/dataset", data, np.float64, "create", True)
    True

    >>> new_data = np.random.rand(5, 5)
    >>> modify_hdf_dataset(Path("output.h5"), "group/dataset", new_data, np.float64, "append", True)
    True

    >>> overwrite_data = np.random.rand(8, 5)
    >>> modify_hdf_dataset(Path("output.h5"), "group/dataset", overwrite_data, np.float64, "overwrite", True)
    True
    """

    # initial shape is zero in first dimension, with no size limit
    shape, maxshape = list(data.shape), list(data.shape)
    if mutable:
        # shape[0], maxshape[0] = 0, None
        maxshape[0] = None
    shape, maxshape = tuple(shape), tuple(maxshape)

    with h5py.File(hdf_path, "r+") as hdf_file:
        # with h5py.File(hdf_path, "r+", locking=False) as hdf_file:
        if operation == "create":
            if not dataset_loc in hdf_file:

                dataset = hdf_file.create_dataset(
                    name=dataset_loc,
                    data=data,
                    shape=shape,
                    dtype=dtype,
                    # compression="gzip", #TODO 4/2/24 compression not allowing read in HDF_View
                    maxshape=maxshape,
                    chunks=True,
                )
                return True
            else:
                raise ValueError(f'Dataset already exists, cannot "{operation}"')

        dataset = hdf_file[dataset_loc]

        if operation == "append":
            # appending to dataset
            dataset.resize(dataset.shape[0] + data.shape[0], axis=0)
            dataset[-data.shape[0] :] = data
            return True

        elif operation == "overwrite":
            # shrinking dataset
            dataset.resize(data.shape[0], axis=0)
            dataset[:] = data
            return True

        else:
            raise KeyError(
                f'operation of "{operation}" requested, only, {"create"}, {"expand"}, and {"overwrite"} currently supported.'
            )


def write_attr_dict(hdf_path: Path, d: dict, dataset_loc: str) -> bool:
    """Write attributes to group or dataset in the h5 file
    from the provided dictionary

    hdf_path : path object to location of hdf file
    d: dict of attributes to add
    dataset_loc: internal path to the dataset in the hdf file (e.g. "file/container/container2/dataset")

    """

    with h5py.File(hdf_path, "r+") as hdf_file:
        loc = hdf_file[dataset_loc]
        # print(d, type(d))
        for key, value in d.items():
            # print(key, value)
            if isinstance(value, Path):
                value = str(value)
            dt = h5py.special_dtype(vlen=str)

            loc.attrs.create(name=key, data=value, dtype=dt)

    return


### INITIAL DATA CREATION ###
def write_h5(hdf_path: Path, semantic_stack: SemanticImageStack) -> bool:
    """Write new h5 file
    the data will be written into a dataset
    the metadata will be written as attributes of the dataset

    hdf_path : path object to location of hdf file
    semantic_stack: the semantic image stack

    """

    base_container = "VoxelData"
    create_hdf(hdf_path=hdf_path, base_container=base_container)

    img_data = semantic_stack.data
    dtype = type(
        img_data.flat[0]
    )  # data.item(0) will return python dtype, which may be better
    dataset_loc = f"{base_container}/{semantic_stack.name}"
    modify_hdf_dataset(
        hdf_path=hdf_path,
        dataset_loc=dataset_loc,
        data=img_data,
        dtype=dtype,
        operation="create",
        mutable=False,
    )

    # TODO 4/2/2024 write metadata as attributes
    md = semantic_stack.metadata
    # TODO 5/7/2024 make metadata strings add to h5
    metadata_dict = ut.metadata_to_dict(md)
    # d = {"height": }
    write_attr_dict(hdf_path=hdf_path, d=metadata_dict, dataset_loc=dataset_loc)


### ALL OTHER FUNCTIONALITIES ###
@singledispatch
def add_to_h5(
    data,
    h5_path: Path,
    h5_group: str,  # internal "folder" in the hdf to save data
):
    """
    Add data to an HDF5 file.

    This is a generic function that adds data to an HDF5 file. Specific implementations
    are provided for different types of data.

    Parameters
    ----------
    data : object
        The data to be added to the HDF5 file.
    h5_path : Path
        The path to the HDF5 file.
    h5_group : str
        The internal "folder" in the HDF5 file to save data.

    Returns
    -------
    None

    Raises
    ------
    NotImplementedError
        If no handler is registered for the type of `data`.

    Examples
    --------
    >>> add_to_h5(some_data, Path("output.h5"), "group_name")
    NotImplementedError: No handler for type <class 'type'>
    """
    _ = data
    __ = h5_path
    ___ = h5_group
    raise NotImplementedError(f"No handler for type {type(data)}")


@add_to_h5.register(Centroids)
def _(
    data,
    h5_path: Path,
    h5_group: str,
):
    """
    Add centroid data to an HDF5 file.

    This function processes and adds centroid data to an HDF5 file.

    Parameters
    ----------
    data : Centroids
        The centroid data to be added.
    h5_path : Path
        The path to the HDF5 file.
    h5_group : str
        The internal "folder" in the HDF5 file to save data.

    Returns
    -------
    None

    Examples
    --------
    >>> centroids = Centroids(...)
    >>> add_to_h5(centroids, Path("output.h5"), "group_name")
    """

    units = data.data[0].cx.unit.value
    centroid_data = ut.centroids_to_ndarray(centroids=data)

    # add data
    dataset_loc = f"{h5_group}/centroids"
    modify_hdf_dataset(
        hdf_path=h5_path,
        dataset_loc=dataset_loc,
        data=centroid_data,
        dtype=float,
        operation="create",
        mutable=False,
    )
    attrs = {
        "nlabels": str(centroid_data.shape[0]),
        "units": str(units),
        "ordering": str("X, Y, Z"),
    }
    write_attr_dict(
        hdf_path=h5_path,
        d=attrs,
        dataset_loc=dataset_loc,
    )


@add_to_h5.register(BestFitEllipsoids)
def _(
    data,
    h5_path: Path,
    h5_group: str,
):
    """
    Add best fit ellipsoid data to an HDF5 file.

    This function processes and adds best fit ellipsoid data to an HDF5 file.

    Parameters
    ----------
    data : BestFitEllipsoids
        The best fit ellipsoid data to be added.
    h5_path : Path
        The path to the HDF5 file.
    h5_group : str
        The internal "folder" in the HDF5 file to save data.

    Returns
    -------
    None

    Examples
    --------
    >>> ellipsoids = BestFitEllipsoids(...)
    >>> add_to_h5(ellipsoids, Path("output.h5"), "group_name")
    """

    axis_units = data.data[0].a.length.unit.value
    axis_lengths, axis_vectors = ut.ellipsoids_to_ndarray(data)

    # add data
    dataset_loc = f"{h5_group}/semi-axis_lengths"
    modify_hdf_dataset(
        hdf_path=h5_path,
        dataset_loc=dataset_loc,
        data=axis_lengths,
        dtype=float,
        operation="create",
        mutable=False,
    )
    attrs = {
        "notes": str("property of best fit ellipsoid"),
        "nlabels": str(axis_lengths.shape[0]),
        "units": str(axis_units),
        "ordering": str("a, b, c, with a > b > c"),
    }
    write_attr_dict(
        hdf_path=h5_path,
        d=attrs,
        dataset_loc=dataset_loc,
    )

    # add orientation data
    dataset_loc = f"{h5_group}/axis_vectors"
    modify_hdf_dataset(
        hdf_path=h5_path,
        dataset_loc=dataset_loc,
        data=axis_vectors,
        dtype=float,
        operation="create",
        mutable=False,
    )
    attrs = {
        "notes": str("property of best fit ellipsoid"),
        "nlabels": str(axis_lengths.shape[0]),
        "units": "unit vector",
        "ordering": str(
            "u, v, w for each axis a, b, c \n\t(a_u, a_v, a_w, b_u, b_v, b_w, c_u, c_v, c_w)"
        ),
    }
    write_attr_dict(
        hdf_path=h5_path,
        d=attrs,
        dataset_loc=dataset_loc,
    )


@add_to_h5.register(EllipsoidSurfaceAreas)
def _(
    data,
    h5_path: Path,
    h5_group: str,
):
    """
    Add ellipsoid surface area data to an HDF5 file.

    This function processes and adds ellipsoid surface area data to an HDF5 file.

    Parameters
    ----------
    data : EllipsoidSurfaceAreas
        The ellipsoid surface area data to be added.
    h5_path : Path
        The path to the HDF5 file.
    h5_group : str
        The internal "folder" in the HDF5 file to save data.

    Returns
    -------
    None

    Examples
    --------
    >>> surface_areas = EllipsoidSurfaceAreas(...)
    >>> add_to_h5(surface_areas, Path("output.h5"), "group_name")
    """

    area_units = data.data[0].unit_squared.value
    surface_areas = ut.surface_areas_to_ndarray(data)

    # add data
    dataset_loc = f"{h5_group}/ellipsoid_surface_areas"
    modify_hdf_dataset(
        hdf_path=h5_path,
        dataset_loc=dataset_loc,
        data=surface_areas,
        dtype=float,
        operation="create",
        mutable=False,
    )
    attrs = {
        "notes": str("property of best fit ellipsoid"),
        "nlabels": str(surface_areas.shape[0]),
        "units_squared": f"{area_units}",
        "method": "Knud Thomsen approximation for scalene ellipsoids (2004)",
    }
    write_attr_dict(
        hdf_path=h5_path,
        d=attrs,
        dataset_loc=dataset_loc,
    )


@add_to_h5.register(EllipsoidVolumes)
def _(
    data,
    h5_path: Path,
    h5_group: str,
):
    """
    Add ellipsoid volume data to an HDF5 file.

    This function processes and adds ellipsoid volume data to an HDF5 file.

    Parameters
    ----------
    data : EllipsoidVolumes
        The ellipsoid volume data to be added.
    h5_path : Path
        The path to the HDF5 file.
    h5_group : str
        The internal "folder" in the HDF5 file to save data.

    Returns
    -------
    None

    Examples
    --------
    >>> volumes = EllipsoidVolumes(...)
    >>> add_to_h5(volumes, Path("output.h5"), "group_name")
    """
    area_units = data.data[0].unit_cubed.value
    ellipsoid_volumes = ut.volumes_to_ndarray(data)

    # add data
    dataset_loc = f"{h5_group}/ellipsoid_volumes"
    modify_hdf_dataset(
        hdf_path=h5_path,
        dataset_loc=dataset_loc,
        data=ellipsoid_volumes,
        dtype=float,
        operation="create",
        mutable=False,
    )
    attrs = {
        "notes": str("property of best fit ellipsoid"),
        "nlabels": str(ellipsoid_volumes.shape[0]),
        "units_cubed": f"{area_units}",
        "method": "4/3 * pi * a * b * c",
    }
    write_attr_dict(
        hdf_path=h5_path,
        d=attrs,
        dataset_loc=dataset_loc,
    )


@add_to_h5.register(InstanceImageStack)
def _(
    data,
    h5_path: Path,
    h5_group: str,
):
    """
    Add instance image stack data to an HDF5 file.

    This function processes and adds instance image stack data to an HDF5 file.

    Parameters
    ----------
    data : InstanceImageStack
        The instance image stack data to be added.
    h5_path : Path
        The path to the HDF5 file.
    h5_group : str
        The internal "folder" in the HDF5 file to save data.

    Returns
    -------
    None

    Examples
    --------
    >>> image_stack = InstanceImageStack(...)
    >>> add_to_h5(image_stack, Path("output.h5"), "group_name")
    """

    # print(f"data type: {type(data)}")

    instance_image_stack = data

    # add data
    dataset_loc = f"{h5_group}/{instance_image_stack.name}"
    modify_hdf_dataset(
        hdf_path=h5_path,
        dataset_loc=dataset_loc,
        data=instance_image_stack.data,
        dtype=type(instance_image_stack.data.flat[0]),
        operation="create",
        mutable=False,
    )

    # add metadata
    md = instance_image_stack.metadata
    metadata_dict = ut.metadata_to_dict(md)
    write_attr_dict(
        hdf_path=h5_path,
        d=metadata_dict,
        dataset_loc=dataset_loc,
    )

    extra_attrs = {
        "nlabels": str(instance_image_stack.nlabels),
        "min_feature_size": str(instance_image_stack.min_feature_size),
    }
    write_attr_dict(
        hdf_path=h5_path,
        d=extra_attrs,
        dataset_loc=dataset_loc,
    )


@add_to_h5.register(InstanceIndices)
def _(
    data,
    h5_path: Path,
    h5_group: str,
):
    """
    Add instance indices data to an HDF5 file.

    This function processes and adds instance indices data to an HDF5 file.

    Parameters
    ----------
    data : InstanceIndices
        The instance indices data to be added.
    h5_path : Path
        The path to the HDF5 file.
    h5_group : str
        The internal "folder" in the HDF5 file to save data.

    Returns
    -------
    None

    Examples
    --------
    >>> indices = InstanceIndices(...)
    >>> add_to_h5(indices, Path("output.h5"), "group_name")
    """
    raise NotImplementedError(
        "Ability to write out instance indices not yet implemented."
    )
    # TODO: AP, write out as single variable length data instead of individual datasets (currently)
    # https://docs.h5py.org/en/stable/special.html
    # dt = h5py.vlen_dtype(np.dtype('int32'))
    instance_indices = data

    base_container = f"{instance_indices.source_name}_indices"
    # NumPy doesn’t support ragged arrays, and the ‘arrays of arrays’
    # h5py uses as a workaround are not as convenient or efficient as
    # regular NumPy arrays. If you’re deciding how to store data,
    # consider whether there’s a sensible way to do it without a
    # variable-length type.
    ragged_label_indices_array = np.array((len(instance_indices.indices)))
    for each_label in instance_indices.labels.data:
        ragged_label_indices_array[each_label] = instance_indices.indices[each_label]

    dataset_loc = f"{base_container}/label_indices"
    modify_hdf_dataset(
        hdf_path=h5_path,
        dataset_loc=dataset_loc,
        data=ragged_label_indices_array,
        dtype=h5py.vlen_dtype(np.dtype("int32")),
        operation="create",
        mutable=False,
    )
    # # Tree the labels
    # for (
    #     each_label
    # ) in (
    #     instance_indices.labels.data
    # ):  # TODO: why does the instance_indices.labels not return the correct type, instead it provides the ndarray...

    #     dataset_loc = f"{base_container}/label_{each_label:06d}"
    #     modify_hdf_dataset(
    #         hdf_path=h5_path,
    #         dataset_loc=dataset_loc,
    #         data=instance_indices.indices[each_label],
    #         dtype=type(instance_indices.indices[each_label].flat[0]),
    #         operation="create",
    #         mutable=False,
    #     )


@add_to_h5.register(InstanceProperties)
def _(
    data,
    h5_path: Path,
    h5_group: str,
):
    """
    Add instance properties data to an HDF5 file.

    This function processes and adds instance properties data to an HDF5 file.

    Parameters
    ----------
    data : InstanceProperties
        The instance properties data to be added.
    h5_path : Path
        The path to the HDF5 file.
    h5_group : str
        The internal "folder" in the HDF5 file to save data.

    Returns
    -------
    None

    Examples
    --------
    >>> properties = InstanceProperties(...)
    >>> add_to_h5(properties, Path("output.h5"), "group_name")
    """

    # centroids
    add_to_h5(data.centroids, h5_path=h5_path, h5_group=h5_group)

    # ellipsoids
    add_to_h5(data.ellipsoids, h5_path=h5_path, h5_group=h5_group)

    # surface area
    add_to_h5(data.surface_areas, h5_path=h5_path, h5_group=h5_group)

    add_to_h5(data.volumes, h5_path=h5_path, h5_group=h5_group)

    # equiv spherical diameter
    eq_diam = data.equivalent_sphere_diameters
    diam = [i.value for i in eq_diam]
    diam_data = np.array(diam, dtype=float).T
    diam_units = eq_diam[0].unit.value

    dataset_loc = f"{h5_group}/equivalent_sphere_diameters"
    modify_hdf_dataset(
        hdf_path=h5_path,
        dataset_loc=dataset_loc,
        data=diam_data,
        dtype=float,
        operation="create",
        mutable=False,
    )
    attrs = {
        "notes": str("from volume detemined by voxel count and resolution"),
        "nlabels": str(diam_data.shape[0]),
        "units": str(diam_units),
    }
    write_attr_dict(
        hdf_path=h5_path,
        d=attrs,
        dataset_loc=dataset_loc,
    )

    # num voxels
    num_voxel = data.n_voxels
    n_voxels = [i.value for i in num_voxel]
    n_vox_data = np.array(n_voxels, dtype=int).T
    n_vox_units = num_voxel[0].unit.value

    dataset_loc = f"{h5_group}/num_voxels"
    modify_hdf_dataset(
        hdf_path=h5_path,
        dataset_loc=dataset_loc,
        data=n_vox_data,
        dtype=int,
        operation="create",
        mutable=False,
    )
    attrs = {
        "nlabels": str(diam_data.shape[0]),
        "units": str(n_vox_units),
    }
    write_attr_dict(
        hdf_path=h5_path,
        d=attrs,
        dataset_loc=dataset_loc,
    )


@add_to_h5.register(NthNearestNeighbors)
def _(
    data,
    h5_path: Path,
    h5_group: str,
):
    """
    Add nth nearest neighbor data to an HDF5 file.

    This function processes and adds nth nearest neighbor data to an HDF5 file.

    Parameters
    ----------
    data : NthNearestNeighbors
        The nth nearest neighbor data to be added.
    h5_path : Path
        The path to the HDF5 file.
    h5_group : str
        The internal "folder" in the HDF5 file to save data.

    Returns
    -------
    None

    Examples
    --------
    >>> neighbors = NthNearestNeighbors(...)
    >>> add_to_h5(neighbors, Path("output.h5"), "group_name")
    """

    aa = 2
    units = data.distances[0].unit.value

    distance_list = data.distances
    distances = [i.value for i in distance_list]
    distance_array = np.array(distances, dtype=float).T
    neighbor_id = data.instance_id

    dataset_loc = f"{h5_group}/nearest_neighbor_distances"
    modify_hdf_dataset(
        hdf_path=h5_path,
        dataset_loc=dataset_loc,
        data=distance_array,
        dtype=float,
        operation="create",
        mutable=False,
    )
    attrs = {
        "nlabels": str(distance_array.shape[0]),
        "units": str(units),
        "nth_nearest": f"{data.nth_nearest}: 1st nearest is the feature itself",
    }
    write_attr_dict(
        hdf_path=h5_path,
        d=attrs,
        dataset_loc=dataset_loc,
    )

    dataset_loc = f"{h5_group}/nearest_neighbor_IDs"
    modify_hdf_dataset(
        hdf_path=h5_path,
        dataset_loc=dataset_loc,
        data=neighbor_id,
        dtype=int,
        operation="create",
        mutable=False,
    )
    attrs = {
        "nlabels": str(neighbor_id.shape[0]),
        "nth_nearest": f"{data.nth_nearest}: 1st nearest is the feature itself",
    }
    write_attr_dict(
        hdf_path=h5_path,
        d=attrs,
        dataset_loc=dataset_loc,
    )


### CONVERT BETWEEN H5 <-> FOLDER OF IMAGES   ###
def image_to_voxel(yml_path: Path) -> Path:
    """
    Populate the HDF5 file with the semantic segmentation image stack specified in the YAML file, including metadata.

    This function reads the YAML file to obtain the necessary parameters, processes the image stack, and writes the resulting data to an HDF5 file.

    Parameters
    ----------
    yml_path : Path
        The path to the YAML file containing the configuration and parameters.

    Returns
    -------
    Path
        The path to the created HDF5 file.

    Raises
    ------
    ValueError
        If the "cli_entry_points" key in the YAML file does not contain "image_to_voxel".

    Examples
    --------
    >>> yml_path = Path("config.yml")
    >>> hdf5_path = image_to_voxel(yml_path)
    Wrote output file: /path/to/output.h5
    >>> print(hdf5_path)
    /path/to/output.h5
    """

    # Import images and metadata into sematic_stack
    yml_vals = ut.yaml_to_dict(yml_path)

    # check cli_entry_points is valid
    if "image_to_voxel" not in yml_vals["cli_entry_points"]:
        raise ValueError(
            f"""Error. Incorrect yml format. 
                This function requires the "cli_entry_points" key to contain "image_to_voxel", 
                but currently contains the following options: {yml_vals["cli_entry_points"]} """
        )
    semantic_stack_save = ia.process_image_stack(yml_path)

    # Write semantic_stack to hdf
    h5_name = yml_vals["h5_filename"] + ".h5"

    # path_file_output = Path(yml_vals["out_dir"]).joinpath(h5_name)
    path_file_output = Path(yml_vals["out_dir"]).expanduser().joinpath(h5_name)
    write_h5(path_file_output, semantic_stack_save)

    print(f"Wrote output file: {path_file_output}")

    return path_file_output


def image_to_voxel_command_line():
    """
    The command line wrapper for the `image_to_voxel` function.

    This function sets up the command line argument parser, parses the input arguments,
    and calls the `image_to_voxel` function with the provided YAML input file.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Examples
    --------
    To run this function from the command line:

        $ python -m your_module_name input_file.yml
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("input_file", help="the .yml user input file")
    args = parser.parse_args()
    input_file = args.input_file

    image_to_voxel(yml_path=input_file)


def voxel_to_image(yml_path: Path) -> Path:
    """
    Save the image data within the HDF5 file as TIFFs in a new directory.

    This function reads the YAML file to obtain the necessary parameters, extracts the image data from the HDF5 file,
    and saves the images as TIFF files in a specified directory.

    Parameters
    ----------
    yml_path : Path
        The path to the YAML file containing the configuration and parameters.

    Returns
    -------
    Path
        The path to the directory containing the saved images.

    Raises
    ------
    ValueError
        If the "cli_entry_points" key in the YAML file does not contain "voxel_to_image".
        If the specified slicing direction is not valid.

    Examples
    --------
    >>> yml_path = Path("config.yml")
    >>> image_dir = voxel_to_image(yml_path)
    >>> print(image_dir)
    /path/to/output/images
    """

    yml_vals = ut.yaml_to_dict(yml_path)

    # check cli_entry_points is valid
    if "voxel_to_image" not in yml_vals["cli_entry_points"]:
        raise ValueError(
            f"""Error. Incorrect yml format. 
            This function requires the "cli_entry_points" key to contain "voxel_to_image", 
            but currently contains the following options: {yml_vals["cli_entry_points"]} """
        )

    hdf_path = Path(yml_vals["voxel_data_path"]).expanduser()
    # TODO add alternative to ingest npy data dir

    hdf_dataset_location = yml_vals["voxel_data_location"]
    output_image_dir = Path(yml_vals["image_parent_dir"]).expanduser()
    output_image_type = yml_vals["image_output_type"]

    slice_normal = yml_vals["image_slice_normal"]
    valid_slice_normal = set(item.name for item in cs.CartesianAxis3D)
    if slice_normal not in valid_slice_normal:
        raise ValueError(
            f"Error, '{slice_normal}' is not a valid slicing direction, accepted units are: {valid_slice_normal}"
        )
    slice_axis = cs.CartesianAxis3D[slice_normal]

    with h5py.File(hdf_path, "r") as f:
        data = np.squeeze(f[hdf_dataset_location][:])
    ut.ndarray_to_img(
        data=data,
        parent_dir=output_image_dir,
        folder_name=Path(hdf_dataset_location).stem,
        file_type=output_image_type,
        slice_axis=slice_axis,
    )

    return output_image_dir.joinpath(hdf_path.stem)


def voxel_to_image_command_line():
    """
    The command line wrapper for the `voxel_to_image` function.

    This function sets up the command line argument parser, parses the input arguments,
    and calls the `voxel_to_image` function with the provided YAML input file.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Examples
    --------
    To run this function from the command line:

        $ python -m your_module_name input_file.yml
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("input_file", help="the .yml user input file")
    args = parser.parse_args()
    input_file = args.input_file

    new_path = voxel_to_image(yml_path=input_file)

    print(f"\nVoxel data extracted to the following relative directory:'{new_path}'")



"""
Command Line Entry Points Module
================================

This module provides command line entry points for various functions
and utilities.  It serves as the interface between the command line
and the underlying functionality of the application.
"""

from typing import Final

# import pkg_resources  # part of setup tools
from importlib.metadata import version
import recon3d.constants as cs

CLI_DOCS: Final[
    str
] = """
--------
recon3d
--------

recon3d
    (this command)

binary_to_semantic <path_to_file>.yml
    Converts binary image stack to semantic image stack in a
    folder specified in the user input .yml file.

    Example:
        # Edit path variables in
        # ~/recon3d/recon3d/examples/binary_to_semantic.yml
        (.venv) recon3d> downscale recon3d/examples/binary_to_semantic.yml

downscale <path_to_file>.yml
    Downscales images in a folder specified in the user input .yml file.

    Example:
        # Edit path variables in
        # ~/recon3d/recon3d/examples/downscale_thunder.yml
        (.venv) recon3d> downscale recon3d/examples/downscale_thunder.yml

hello
    Prints 'Hello world!' to the terminal to illustrate command line
    entry points.

image_to_voxel <path_to_file>.yml
    From a single image (or image stack) in a folder specified in the
    user input .yml file, creates a hdf file in the specified
    output folder.

    Example:
        # Edit path variables in
        # ~/recon3d/recon3d/examples/image_to_voxel.yml
        (.venv) recon3d> image_to_voxel recon3d/examples/image_to_voxel.yml

instance_analysis <path_to_file>.yml
    Digest a semantic segmention accessible as a folder containing an image
    stack specified in the user input .yml file.

    Example:
        # Edit path variables in
        # ~/recon3d/recon3d/examples/instance_analysis.yml
        (.venv) recon3d> instance_analysis recon3d/examples/instance_analysis.yml

npy_to_mesh <path_to_file>.yml
    Converts an instance or semantic segmentation, encoded as a .npy file,
    to an Exodus II finite element mesh using automesh.
    See https://autotwin.github.io/automesh/

    Example:
        # Edit path variables in
        # ~/recon3d/tests/examples/letter_f_3d.yml
        (.venv) recon3d> npy_to_mesh tests/examples/letter_f_3d.yml

semantic_to_binary <path_to_file>.yml
    Converts semantic image stack to series of binary image stacks in
    a folder specified in the user input .yml file

    Example:
        # Edit path variables in
        # ~/recon3d/tests/examples/semantic_to_binary.yml
        (.venv) recon3d> downscale recon3d/examples/semantic_to_binary.yml

void_descriptor <path_to_file>.yml
    From a pore dataset contained within a hdf file specified by
    the input .yml file, compute the void descriptor attributes
    for the void descriptor function.

    Example:
        # Edit path variables in
        # ~/recon3d/recon3d/examples/void_descriptor.yml
        (.venv) recon3d> void_descriptor recon3d/examples/void_descriptor.yml

voxel_to_image <path_to_file>.yml
    From a dataset contained within a hdf file specified by the input
    .yml file, creates an image stack with the same dataset name in
    the specified parent output folder.

    Example:
        # Edit path variables in
        # ~/recon3d/docs/userguide/src/voxel_to_mesh/letter_f_3d.yml
        (.venv) recon3d> image_to_voxel docs/userguide/src/voxel_to_mesh/letter_f_3d.yml

"""


def recon3d():
    """
    Prints the command line documentation to the command window.

    This function prints the contents of the global variable `CLI_DOCS` to the
    command window. It is assumed that `CLI_DOCS` contains the necessary
    documentation in string format.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """
    print(CLI_DOCS)


def hello() -> str:
    """
    Simple example of a function hooked to a command line entry point.

    This function serves as an example of how to hook a function to a command line
    entry point. When called, it returns the canonical "Hello world!" string.

    Parameters
    ----------
    None

    Returns
    -------
    str
        The canonical "Hello world!" string.
    """

    return "Hello world!"


def module_version():
    """
    Prints the module version and the yml_schema_version.

    This function retrieves the version of the module and the YAML schema version
    from the `Constants` class in the `cs` module. It prints these versions to the
    command window and returns the module version.

    Parameters
    ----------
    None

    Returns
    -------
    ver : str
        The version of the module.
    """

    # ver = pkg_resources.require("recon3d")[0].version
    module_name = cs.Constants().module_short_name
    ver = version(module_name)
    print(f"module version: {ver}")
    print(f"  yml_schema_version: {cs.Constants().YML_SCHEMA_VERSION}")
    return ver


def work_in_progress():
    """
    Prints the 'Work in Progress (WIP)' warning message.

    This function prints a warning message indicating that the function is a
    work in progress and has not yet been implemented.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """
    print("Warning: Work in progress (WIP), function not yet implemented.")



"""This module tests individual functions for the downscale module."""

# python standard libraries
from pathlib import Path

# import shutil # unused import

# 3rd party libraries
import pytest

# import yaml

# local libraries
from recon3d import downscale as ds
import recon3d.utility as ut
from recon3d.static_test_paths import *

# @pytest.fixture
# def test_file() -> Path:
#     """The relative path and file string locating the default yml test file."""

#     return Path(__file__).parent.joinpath("examples", "downscale_example.yml")


def test_downscale():
    """Tests that the test images are downscaled."""
    aa = DOWNSCALE_YML
    db = ut.yaml_to_dict(aa)

    # overwrite so that this test runs on any machine
    db["image_dir"] = str(
        Path(__file__).parent.joinpath("data", "cylinder_machined_semantic")
    )
    db["out_dir"] = str(
        Path(__file__).parent.joinpath("test_output", "cylinder_machined")
    )

    temp_path = Path(__file__).parent.joinpath("test_output", "temp.yml")
    temp_path.parent.mkdir(
        parents=True, exist_ok=True
    )  # makes directory if it doesn't exist
    # make sure the recon3d/examples/temp.yml doesn't already
    # exist from a previous test run or from a test paused
    # midway through
    try:
        temp_path.unlink()  # delete the temp.yml if it exists aready
    except FileNotFoundError:
        pass

    bb = ut.dict_to_yaml(db, str(temp_path))
    success = ds.downscale(path_file_input=str(bb))

    ut.rmdir(temp_path.parent)
    # temp_path.parent.rmdir()  # delete the temp.yml after checking

    assert success


def test_apply_bbox():
    """
    Test the bounding box is working"""
    dataset_path = Path(__file__).parent.joinpath("data", "cylinder_machined_semantic")
    image_type = ".tif"

    image_stack = ut.read_images(dataset_path, image_type)

    threshold = 0
    bb_image_stack = ds.apply_bbox(image_stack, threshold)

    correct_shape = (21, 144, 145)

    assert bb_image_stack.shape == correct_shape


def test_bbox_range():
    """test it calculates the bounding box range correctly"""
    dataset_path = Path(__file__).parent.joinpath("data", "cylinder_machined_semantic")
    image_type = ".tif"

    image_stack = ut.read_images(dataset_path, image_type)
    threshold = 0
    found_bbox_range = ds.bbox_range(image_stack, threshold)

    known_bbox_range = (0, 20, 3, 146, 4, 148)

    assert found_bbox_range == known_bbox_range


def test_downscale_save_stack():
    """check the cropped images are saved"""
    dataset_path = Path(__file__).parent.joinpath("data", "cylinder_machined_semantic")
    image_type = ".tif"

    image_stack = ut.read_images(dataset_path, image_type)
    save_path = Path(__file__).parent.joinpath("test_output", "cylinder_machined")
    folder_suffix = f"{int(100.0)}_dx"

    ds.save_downscale_stack(image_stack, save_path, folder_suffix)

    known_output_folder_path = save_path.joinpath(
        f"images_at_resolution_{folder_suffix}"
    )
    found_image_stack = ut.read_images(known_output_folder_path, image_type)

    # # delete the output files after checking
    # try:
    #     shutil.rmtree(save_path)
    # except OSError as e:
    #     print(f"Failed to delete directory: {e}")

    ut.rmdir(Path(__file__).parent.joinpath("test_output"))

    assert found_image_stack.shape == image_stack.shape


# move to test_utility.py
#
# def test_dict_to_yaml():
#     """Tests that the dict is converted to a yaml file."""
#     temp_path = Path(__file__).parent.joinpath("files", "temp.yml")
#     db = {
#         "image_dir": None,
#         "image_type": ".tif",
#         "resolution_input": {"dx": 10.0, "dy": 10.0, "dz": 10.0},
#         "resolution_output": {"dx": 100.0, "dy": 100.0, "dz": 10.0},
#         "downscale_tolerance": 0.0001,
#         "image_limit_factor": 2.0,
#         "out_dir": None,
#         "padding": {"nx": 1, "ny": 1, "nz": 1},
#         "output_stack_type": "padded",
#         "save_npy": True,
#         "writeVTR": True,
#     }
#     temp_yaml = ds.dict_to_yaml(db, str(temp_path))
#     found_yaml = Path(__file__).parent.joinpath("files", "downscale_example.yml")

#     assert ut.compare_files(temp_yaml, found_yaml, [])


# def test_yaml_to_dict(test_file):
#     """Tests that the yaml file is converted to the dict."""
#     aa = test_file
#     known_db = {
#         "image_dir": None,
#         "image_type": ".tif",
#         "resolution_input": {"dx": 10.0, "dy": 10.0, "dz": 10.0},
#         "resolution_output": {"dx": 100.0, "dy": 100.0, "dz": 10.0},
#         "downscale_tolerance": 0.0001,
#         "image_limit_factor": 2.0,
#         "out_dir": None,
#         "padding": {"nx": 1, "ny": 1, "nz": 1},
#         "output_stack_type": "padded",
#         "save_npy": True,
#         "writeVTR": True,
#     }
#     found_db = ds.yaml_to_dict(aa)

#     assert known_db == found_db


def test_padded_dim_small():
    """test padding with a small image size"""
    img_stack_dim_size = 10
    target_res = 7.0
    original_res = 3.0
    tolerance = 0.01
    limit_factor = 2
    found_padded_size = ds.padded_size(
        img_stack_dim_size, target_res, original_res, tolerance, limit_factor
    )

    known_padded_size = 14

    assert found_padded_size == known_padded_size


def test_padded_dim_large():
    """test padding with a small image size"""
    img_stack_dim_size = 15353  # dim2 14192
    target_res = 7.56
    original_res = 1.08
    tolerance = 0.001
    limit_factor = 2
    found_padded_size = ds.padded_size(
        img_stack_dim_size, target_res, original_res, tolerance, limit_factor
    )

    known_padded_size = 15358

    assert found_padded_size == known_padded_size


def test_pad_amount():
    """test padding with a large image size"""
    img_stack_dim_size = 15353  # dim2 14192
    target_res = 7.56
    original_res = 1.08
    tolerance = 0.001
    limit_factor = 2
    found_pad_dim = ds.pad_amount(
        img_stack_dim_size, target_res, original_res, tolerance, limit_factor
    )

    known_pad_dim = (3, 2)  # known dimension, 15358 is divisible

    assert found_pad_dim == known_pad_dim



from pathlib import Path

RECON3D_PATH = Path(__file__).parent.parent.parent

BINARY_TO_SEMANTIC_YML = RECON3D_PATH.joinpath(
    "docs", "userguide", "src", "binary_to_semantic", "binary_to_semantic.yml"
)

DOWNSCALE_YML = RECON3D_PATH.joinpath(
    "docs", "userguide", "src", "downscale", "downscale.yml"
)

IMAGE_TO_VOXEL_YML = RECON3D_PATH.joinpath(
    "docs", "userguide", "src", "voxel_to_image", "image_to_voxel.yml"
)
INSTANCE_ANALYSIS_YML = RECON3D_PATH.joinpath(
    "docs", "userguide", "src", "instance_analysis", "instance_analysis.yml"
)

LETTER_F_3D_EXO = RECON3D_PATH.joinpath(
    "docs", "userguide", "src", "voxel_to_mesh", "letter_f_3d.exo"
)

LETTER_F_3D_YML = RECON3D_PATH.joinpath(
    "docs", "userguide", "src", "voxel_to_mesh", "letter_f_3d.yml"
)

SEMANTIC_TO_BINARY_YML = RECON3D_PATH.joinpath(
    "docs", "userguide", "src", "binary_to_semantic", "semantic_to_binary.yml"
)

VOXEL_TO_IMAGE_YML = RECON3D_PATH.joinpath(
    "docs", "userguide", "src", "voxel_to_image", "voxel_to_image.yml"
)



"""Tests functionality provided by hdf_io.py."""

from pathlib import Path

# Python standard libarires
import glob
from pathlib import Path

# Third party libraries
import numpy as np
import pytest
import h5py

# local libraries
# import recon3d.feature_analysis as fa
import recon3d.constants as cs
import recon3d.instance_analysis as ia
import recon3d.hdf_io as hio
import recon3d.utility as ut
from recon3d.types import *

from tests.test_instance_analysis import create_instance_stack
from recon3d.static_test_paths import *


# @pytest.fixture
# def test_letter_f() -> Path:
#     """Returns the relative path to the letter_f example .tif files."""
#     return Path(__file__).parent.joinpath("data", "letter_f")


# @pytest.fixture
# def test_instance_analysis_yml() -> Path:
#     """The relative path and file string locating the default yml test file."""

#     return Path(__file__).parent.joinpath("examples", "instance_analysis.yml")


### BASE/HELPER FUNCTIONS ###
def test_create_hdf():
    """Test hdf file is created from create_hdf function"""

    hdf_path = Path(__file__).parent.joinpath("data", "test_hdf_io.h5")

    base_container = "test_container"

    hio.create_hdf(hdf_path, base_container)

    assert hdf_path.is_file()

    # remove at end of test
    hdf_path.unlink()


def test_modify_hdf():
    """Test hdf file is created from modify_hdf function"""

    hdf_path = Path(__file__).parent.joinpath("data", "test_hdf_io.h5")

    try:
        hdf_path.unlink()  # delete the test_hdf_io.h5 if it exists aready
    except FileNotFoundError:
        pass

    base_container = "test_container"

    hio.create_hdf(hdf_path, base_container)

    dataset_loc = base_container + "/rand_test_data"
    test_data = np.random.rand(10, 30, 5)
    test_data_dtype = np.float64
    operation = "create"
    mutable = True
    hio.modify_hdf_dataset(
        hdf_path, dataset_loc, test_data, test_data_dtype, operation, mutable
    )

    with h5py.File(hdf_path, "r") as hdf_file:

        found_dataset = hdf_file[dataset_loc]
        assert found_dataset.shape == (10, 30, 5)
        assert found_dataset.dtype == test_data_dtype

    test_data = np.random.rand(12, 30, 5)
    test_data_dtype = np.float64
    operation = "append"
    mutable = True
    hio.modify_hdf_dataset(
        hdf_path, dataset_loc, test_data, test_data_dtype, operation, mutable
    )

    with h5py.File(hdf_path, "r") as hdf_file:

        found_dataset = hdf_file[dataset_loc]
        assert found_dataset.shape == (22, 30, 5)
        assert found_dataset.dtype == test_data_dtype

    operation = "overwrite"
    mutable = True
    hio.modify_hdf_dataset(
        hdf_path, dataset_loc, test_data, test_data_dtype, operation, mutable
    )

    with h5py.File(hdf_path, "r") as hdf_file:

        found_dataset = hdf_file[dataset_loc]
        assert found_dataset.shape == (12, 30, 5)
        assert found_dataset.dtype == test_data_dtype

    # remove at end of test
    hdf_path.unlink()


def test_write_attr_dict():
    """Test attrubutes of dataset in hdf file are added
    from write_attr_dict function"""

    hdf_path = Path(__file__).parent.joinpath("data", "test_hdf_io.h5")

    try:
        hdf_path.unlink()  # delete the test_hdf_io.h5 if it exists aready
    except FileNotFoundError:
        pass

    base_container = "test_container"

    hio.create_hdf(hdf_path, base_container)

    dataset_loc = base_container + "/rand_test_data"
    test_data = np.random.rand(10, 30, 5)
    test_data_dtype = np.float64
    operation = "create"
    mutable = True
    hio.modify_hdf_dataset(
        hdf_path, dataset_loc, test_data, test_data_dtype, operation, mutable
    )

    attr_dict = {"attr1": "dog", "attr2": "cat"}

    hio.write_attr_dict(hdf_path, attr_dict, dataset_loc)

    with h5py.File(hdf_path, "r") as hdf_file:

        found_attrs_keys = dict(hdf_file[dataset_loc].attrs.items())

        assert attr_dict == found_attrs_keys

    # remove at end of test
    hdf_path.unlink()


### ALL OTHER FUNCTIONALITIES ###
def test_add_to_h5():
    """Check the overloaded function can write the different types to the h5 correctly"""

    test_instance_analysis_yml = INSTANCE_ANALYSIS_YML

    output_voxeldata_path = hio.image_to_voxel(test_instance_analysis_yml)

    # InstanceImageStack
    instance_name = "air"
    instance_stack = create_instance_stack(instance_name)
    hio.add_to_h5(
        instance_stack,
        h5_path=output_voxeldata_path,
        h5_group="VoxelData",
    )

    instance_name = "metal"
    instance_stack = create_instance_stack(instance_name)
    hio.add_to_h5(
        instance_stack,
        h5_path=output_voxeldata_path,
        h5_group="VoxelData",
    )

    instance_name = "pore"
    instance_stack = create_instance_stack(instance_name)

    # get instance_indices
    instance_indices = ia.instance_indices(
        instance_stack=instance_stack,
    )
    if instance_stack.min_feature_size > 0:
        # overwrites instance_stack and inst_indices
        instance_stack, instance_indices = ia.minimum_size_filter(
            initial_stack=instance_stack,
            initial_indices=instance_indices,
        )
    hio.add_to_h5(
        instance_stack,
        h5_path=output_voxeldata_path,
        h5_group="VoxelData",
    )

    # Instance dtypes
    # instance_indices = ia.instance_indices(instance_stack)
    # # Centroids
    # centroid_1 = ia.center_of_mass(
    #     indices=instance_indices.indices[2],
    #     resolution=instance_stack.metadata.resolution,
    #     origin=instance_stack.metadata.origin,
    # )
    # centroid_2 = ia.center_of_mass(
    #     indices=instance_indices.indices[3],
    #     resolution=instance_stack.metadata.resolution,
    #     origin=instance_stack.metadata.origin,
    # )
    # centroids = Centroids(data=[centroid_1, centroid_2])
    # hio.add_to_h5(
    #     centroids,
    #     h5_path=output_voxeldata_path,
    #     h5_group="PoreData",
    # )
    # # Ellipsoids
    # ellipsoid_1 = ia.fit_ellipsoid(
    #     indices=instance_indices.indices[2],
    #     centroid=centroid_1,
    #     resolution=instance_stack.metadata.resolution,
    #     origin=instance_stack.metadata.origin,
    # )
    # ellipsoid_2 = ia.fit_ellipsoid(
    #     indices=instance_indices.indices[3],
    #     centroid=centroid_2,
    #     resolution=instance_stack.metadata.resolution,
    #     origin=instance_stack.metadata.origin,
    # )
    # ellipsoids = BestFitEllipsoids(data=[ellipsoid_1, ellipsoid_2])
    # hio.add_to_h5(
    #     ellipsoids,
    #     h5_path=output_voxeldata_path,
    #     h5_group="PoreData",
    # )

    instance_props = ia.instance_properties(
        instance_stack=instance_stack,
        inst_indices=instance_indices,
    )
    hio.add_to_h5(
        instance_props,
        h5_path=output_voxeldata_path,
        h5_group=instance_props.source_name,
    )

    nearest_neighbors = ia.nearest_neighbor_distance(centroids=instance_props.centroids)
    hio.add_to_h5(
        nearest_neighbors,
        h5_path=output_voxeldata_path,
        h5_group=instance_props.source_name,
    )

    # # InstanceIndices
    # instance_indices = ia.instance_indices(instance_stack)
    # hio.add_to_h5(instance_indices, h5_path=output_voxeldata_path)

    output_voxeldata_path.unlink()


### CONVERT BETWEEN H5 <-> FOLDER OF IMAGES   ###
def test_image_to_voxel():
    """Check the images can be properly added to the h5"""

    input_yml = VOXEL_TO_IMAGE_YML  # Path(__file__).parent.joinpath("examples", "voxel_to_image.yml")

    with pytest.raises(ValueError) as err:
        output_voxeldata_path = hio.image_to_voxel(input_yml)
    err_msg = 'Error. Incorrect yml format. \n                This function requires the "cli_entry_points" key to contain "image_to_voxel", \n                but currently contains the following options: [\'voxel_to_image\'] '
    assert err.type == ValueError
    assert err.value.args[0] == err_msg

    # overwrite
    input_yml = IMAGE_TO_VOXEL_YML  # Path(__file__).parent.joinpath("examples", "image_to_voxel.yml")

    output_voxeldata_path = hio.image_to_voxel(input_yml)

    with h5py.File(output_voxeldata_path, "r") as hdf_file:

        found_dataset = hdf_file["/VoxelData/letter_f_test"]
        assert found_dataset.shape == (4, 5, 3)

    # remove at end of test
    output_voxeldata_path.unlink()

    """Check the images from the h5 file can be properly extracted"""

    input_yml = IMAGE_TO_VOXEL_YML  # Path(__file__).parent.joinpath("examples", "image_to_voxel.yml")

    with pytest.raises(ValueError) as err:
        hio.voxel_to_image(input_yml)
    err_msg = 'Error. Incorrect yml format. \n            This function requires the "cli_entry_points" key to contain "voxel_to_image", \n            but currently contains the following options: [\'image_to_voxel\'] '
    assert err.type == ValueError
    assert err.value.args[0] == err_msg

    input_yml = VOXEL_TO_IMAGE_YML  # Path(__file__).parent.joinpath("examples", "voxel_to_image.yml")

    output_images_path = hio.voxel_to_image(input_yml)

    # There should be 4 images written out
    assert len(list(output_images_path.rglob("*.tif"))) == 21

    # remove at end of test
    ut.rmdir(output_images_path)
    ut.rmdir(output_images_path.parent)



""""This module tests the command_line services.

Example:
    To run
    - activate the virutal environment, then
    $ cd ~/recond3d

    pytest (without code coverage)
    pytest tests/test_command_line.py -v  # -v is for verbose

    to run just a single test in this module, for example
    pytest recon3d/tests/test_command_line.py::test_hello_world -v
"""

# from recon3d import command_line as cl
from recon3d import command_line as cl
from recon3d import constants as cs
from importlib.metadata import version


def test_hello_world():
    """Tests that the string 'Hello world!' is returned."""
    print("test_hello_world...")
    known_string_fiducial = "Hello world!"

    found_string = cl.hello()

    assert found_string == known_string_fiducial


def test_version():
    """Tests that the module version is returned."""

    # TODO no package metadata at the moment
    # assert False
    known_version = version(cs.Constants().module_short_name)

    found_version = cl.module_version()
    assert found_version == known_version



"""This module tests the utility module.
"""

# python standard libraries
from pathlib import Path
import numpy as np

# third-party libraries
import pytest

# local libraries
import recon3d.types as cs

# from recon3d import downscale as ds
import recon3d.utility as ut

from recon3d.static_test_paths import *

# @pytest.fixture
# def test_file() -> Path:
#     """The relative path and file string locating the default yml test file."""

#     return Path(__file__).parent.joinpath("examples", "downscale_example.yml")


# @pytest.fixture
# def test_dir() -> Path:
#     """Returns the relative path to the temp output dir for tests"""
#     return Path(__file__).parent.joinpath("files", "outputults")


# @pytest.fixture
# def test_binary_to_semantic_yml() -> Path:
#     """The relative path and file string locating the default yml test file."""

#     return Path(__file__).parent.joinpath("examples", "binary_to_semantic.yml")


# @pytest.fixture
# def test_semantic_to_binary_yml() -> Path:
#     """The relative path and file string locating the default yml test file."""

#     return Path(__file__).parent.joinpath("examples", "semantic_to_binary.yml")


@pytest.fixture
def test_data() -> Path:
    """The relative path and file string locating the default yml test file."""

    return Path(__file__).parent.joinpath("data", "cylinder_machined_semantic")


def create_binary_data(test_data) -> Path:
    """create binary data required for other tests

    Returns:
        path to folder of tif images

    """
    input_data = ut.read_images(test_data)

    outputdir_basepath = Path(__file__).parent.joinpath("data")
    outputdir_name = "binarized_cylinder_machined"

    bw_data = ut.binarize(input_data, 1)

    ut.ndarray_to_img(
        data=bw_data,
        slice_axis=cs.CartesianAxis3D.Z,
        parent_dir=outputdir_basepath,
        folder_name=outputdir_name,
    )

    return outputdir_basepath.joinpath(outputdir_name)


def test_binarize(test_data):
    """Test that semantic image data can be converted to binary image stack"""

    input_data = ut.read_images(test_data)

    # air = 0
    found_binary_data = ut.binarize(input_data, 0)
    found_unique, found_counts = np.unique(found_binary_data, return_counts=True)

    assert found_binary_data.shape == (21, 150, 150)
    assert np.allclose(found_unique, [0, 1])
    assert np.allclose(found_counts, [343817, 128683])

    # metal = 1
    found_binary_data = ut.binarize(input_data, 1)
    found_unique, found_counts = np.unique(found_binary_data, return_counts=True)

    assert found_binary_data.shape == (21, 150, 150)
    assert np.allclose(found_unique, [0, 1])
    assert np.allclose(found_counts, [129114, 343386])

    # pore = 2
    found_binary_data = ut.binarize(input_data, 2)
    found_unique, found_counts = np.unique(found_binary_data, return_counts=True)

    assert found_binary_data.shape == (21, 150, 150)
    assert np.allclose(found_unique, [0, 1])
    assert np.allclose(found_counts, [472069, 431])


def test_binary_with_pores_to_semantic(test_data):
    """Test that binary data can be converted to semantic image stack"""

    # Start with semantic stack, generate a binary stack of the metal
    binary_data_path = create_binary_data(test_data)

    # temporary directory for the semantic stack that will be generated
    semantic_data_path = binary_data_path.parent.joinpath("semantic_image_stack")

    # run function
    found_dict = ut.binary_with_pores_to_semantic(
        input_path=binary_data_path,
        output_path=semantic_data_path,
    )

    print(found_dict)

    found_data = ut.read_images(semantic_data_path)
    assert found_data.shape == (21, 150, 150)

    found_unique, found_counts = np.unique(found_data.data, return_counts=True)

    # check the existing semantic segmentation
    known_data = ut.read_images(test_data)
    known_unique, known_counts = np.unique(known_data, return_counts=True)

    assert np.allclose(found_unique, known_unique)
    assert np.allclose(found_counts, known_counts)

    # check with known value
    assert np.allclose(found_unique, [0, 1, 2])
    assert np.allclose(found_counts, [128683, 343386, 431])

    # clean up
    ut.rmdir(binary_data_path)
    ut.rmdir(semantic_data_path)


def test_binary_to_semantic():
    """Test that a yml can generate semantic image stack from binary image stack"""

    ut.binary_to_semantic(BINARY_TO_SEMANTIC_YML)

    yml_vals = ut.yaml_to_dict(BINARY_TO_SEMANTIC_YML)

    out_dir = Path(yml_vals["out_dir"]).expanduser()

    found_files = list(out_dir.iterdir())

    assert len(found_files) == 21

    ut.rmdir(out_dir)


def test_semantic_to_binary():
    """Test that a yml can generate semantic image stack from binary image stack"""

    ut.semantic_to_binary(SEMANTIC_TO_BINARY_YML)

    yml_vals = ut.yaml_to_dict(SEMANTIC_TO_BINARY_YML)

    out_dir = Path(yml_vals["out_dir"]).expanduser()

    found_files = list(out_dir.iterdir())

    assert len(found_files) == 21

    ut.rmdir(out_dir)


def test_dict_to_yaml():
    """Tests that the dict is converted to a yaml file."""
    temp_path = Path(__file__).parent.joinpath("data", "temp.yml")
    db = {
        "cli_entry_points": ["downscale"],
        "downscale_tolerance": 0.0001,
        "image_dir": None,
        "image_limit_factor": 2.0,
        "image_type": ".tif",
        "out_dir": None,
        "output_stack_type": "padded",
        "padding": {"nx": 1, "ny": 1, "nz": 1},
        "resolution_input": {"dx": 10.0, "dy": 10.0, "dz": 10.0},
        "resolution_output": {"dx": 100.0, "dy": 100.0, "dz": 10.0},
        "save_npy": True,
        "writeVTR": True,
    }
    temp_yaml = ut.dict_to_yaml(db, str(temp_path))
    found_yaml = DOWNSCALE_YML  # Path(__file__).parent.joinpath("examples", "downscale_example.yml")

    assert ut.compare_files(temp_yaml, found_yaml, [])

    try:
        temp_path.unlink()  # delete the temp.yml if it exists aready
    except FileNotFoundError:
        pass


def test_hdf_to_instance_properties():
    """Tests that hdf file can be read to construct InstanceProperties"""

    dataset_path = dataset_path = (
        Path(__file__).parent.joinpath("data", "machined_cylinder.h5").expanduser()
    )

    found_props = ut.hdf_to_instance_properties(
        hdf_path=dataset_path, group_path="pores"
    )

    # TODO more testing with assertions
    assert found_props.ellipsoids.data[4].a.orientation.u == pytest.approx(-0.159112797)
    assert found_props.centroids.data[2].cy.value == pytest.approx(108.8947)


def test_nd_to_img():
    """Tests that the images are saved from a ndarray"""

    # TODO check other packages will work with small letter_f example images
    dataset_path = (
        Path(__file__)
        .parent.joinpath("data", "cylinder_machined_semantic")
        .expanduser()
    )
    img_save_path = (
        Path(__file__).parent.joinpath("output_results", "temp_slices").expanduser()
    )

    # remove at beginning of test
    ut.rmdir(img_save_path)

    image_stack = ut.read_images(dataset_path)

    ut.ndarray_to_img(
        data=image_stack,
        slice_axis=cs.CartesianAxis3D.Z,
        parent_dir=img_save_path,
        folder_name="XY_slices",
    )

    # There should be 4 images written out
    assert len(list(img_save_path.rglob("*.tif"))) == 21

    # remove at end of test
    ut.rmdir(img_save_path)

    ut.ndarray_to_img(
        data=image_stack,
        slice_axis=cs.CartesianAxis3D.X,
        parent_dir=img_save_path,
        folder_name="YZ_slices",
    )

    # There should be 3 more images written out
    assert len(list(img_save_path.rglob("*.tif"))) == 150

    # remove at end of test
    ut.rmdir(img_save_path)

    ut.ndarray_to_img(
        data=image_stack,
        slice_axis=cs.CartesianAxis3D.Y,
        parent_dir=img_save_path,
        folder_name="XZ_slices",
    )

    # There should be 5 more images written out
    assert len(list(img_save_path.rglob("*.tif"))) == 150

    # remove at end of test
    ut.rmdir(img_save_path)


def test_read_images():
    """Tests the image stack is read in."""

    # Test with two images
    dataset_path = (
        Path(__file__)
        .parent.joinpath("data", "cylinder_machined_semantic")
        .expanduser()
    )
    image_type = ".tif"

    image_stack = ut.read_images(dataset_path, image_type)

    correct_shape = (21, 150, 150)

    assert image_stack.shape == correct_shape

    # Test with single image
    dataset_path = Path(__file__).parent.joinpath("data", "thunder_gray")

    image_stack = ut.read_images(dataset_path, image_type)

    correct_shape = (1, 348, 734)

    assert image_stack.shape == correct_shape


def test_difference():
    """Given two tuples with overlapping items, tests than the
    returned tuple is the difference of the first tuple less
    the second tuple.
    """
    x1 = (
        1,
        2,
        3,
        4,
    )
    x2 = (
        4,
        2,
        5,
        6,
    )

    y = (
        1,
        3,
    )

    fx = ut.in_a_but_not_in_b(a=x1, b=x2)

    assert y == fx


def test_pairwise():
    """Tests the pairwise function with a simple test of four items."""
    x = "ABCD"
    y = (("A", "B"), ("B", "C"), ("C", "D"))
    fx = ut.pairwise(x)
    fx2 = tuple(fx)
    assert y == fx2


def test_pairwise_circular():
    """Tests pairwise circular with a simple test of four items."""
    x = "ABCD"
    y = (("A", "B"), ("B", "C"), ("C", "D"), ("D", "A"))
    fx = ut.pairwise_circular(x)
    fx2 = tuple(fx)
    assert y == fx2


def test_yaml_to_dict():
    """Tests that the yaml file is converted to the dict."""
    aa = DOWNSCALE_YML
    known_db = {
        "cli_entry_points": ["downscale"],
        "downscale_tolerance": 0.0001,
        "image_dir": None,
        "image_limit_factor": 2.0,
        "image_type": ".tif",
        "out_dir": None,
        "output_stack_type": "padded",
        "padding": {"nx": 1, "ny": 1, "nz": 1},
        "resolution_input": {"dx": 10.0, "dy": 10.0, "dz": 10.0},
        "resolution_output": {"dx": 100.0, "dy": 100.0, "dz": 10.0},
        "save_npy": True,
        "writeVTR": True,
    }
    found_db = ut.yaml_to_dict(aa)

    assert known_db == found_db
